Artifact: Maximizing System Utilization via Parallelism Management for Co-Located Parallel Applications
=======================================================================================================

Overview
--------------------------------------------------------------------------------

This package contains the artifact to evaluate the NuPoCo framework and to reproduce the results discussed in the paper on an Intel Xeon machine.
The artifact consists of three components.

**Benchmark suite (the benchmark directory)**: it includes the source code of benchmark applications and several scripts to automate the experiments.

**Runtime environment (the rte directory)**: it includes the NuPoCo resource manager (the performance model and the parallelism manager) and
a library that connects between the resource manager and the OpenMP applications.

**The GOMP library (the libgomp directory)**: several different versions of the GOMP library are provided.
One of them implements the cooperative work scheduling technique.

Installation
--------------------------------------------------------------------------------

### Access to the 72-core Intel Machine
If necessary, we may provide access to the hardware used to produce the published results. Send an email to bernhard@csap.snu.ac.kr to receive login credentials for the Intel machine. 

### The runtime environment (RTE)

The code for the runtime environment is located in the *rte* directory.
Some parts of the code are provided in binary form.

The installation script

```
$ source ./env.sh
$ sh install.sh
```

The *install.sh* script consists of the following two steps:
1. setting RTE path and compiling (the PCM library has already been installed):

```
$ cd rte
$ source env.sh
$ make pcm -j  (for Intel system only)
$ make -j
```

2. building the communication library

```
$ cd rte/rte_comm/
$ make
```

This will create four different deamon files, **nupoco_monitor**, **greedy_monitor**, **c3po_monitor**, **eqp_monitor**, to support different execution modes in the evaluation.

### The GOMP library

The artifact contains the pre-compiled library files in the *libgomp* directory.
There are three different pre-compiled GOMP libraries to support the execution modes described in Section 7.3.

**libgomp-5.4.0-original**: the unmodified GOMP library used by the **Native** mode. Here, thread pinning is disabled to allow the Linux scheduler to perform thread and data placement.

**libgomp-5.4.0-binding**: the GOMP library where threads are pinned to each core. This library is used in **Batch** mode since it provides the best performance with the **Batch** scheduling method.

**libgomp-5.4.0-coop**: the cooperative work scheduling (Section 6) is applied to adjust the number of threads of the applications. **NuPoCo** uses this library.

If you want to install the libraries step by step, refer to the README file in the *libgomp* directory.

### The benchmark suite

The benchmark applications are located in the *benchmark/applications* directory.

Installation script

```
$ sh install.sh
```

Benchmark: Reproducing the results
--------------------------------------------------------------------------------

We provide the following scripts to conduct the experiments for the co-location scenarios described in Section 7.4.
Note that root privileges are required to access the performance monitoring units.

```
$ sudo su
```

First, execute one of the provided configuration files to select one of the execution modes,
**Batch**, **Native**, **EQP**, **Scalability**, **NuPoCo Greedy** and **NuPoCo**
as described in Section 7.3.
Then, run *run\_benchmark\_standalone.py*.
The script will save the log files into the *log* directory.

Note that testing one configuration takes about one hour.

### Batch

```
# source config_standalone.sh
# python run_benchmark_standalone.py
```

### Native

```
# source config_native.sh
# python run_benchmark_scenario.py
```

### EQP

```
# source config_eqp.sh
# python run_benchmark_scenario.py
```

### Scalability

```
# source config_c3po.sh
# python run_benchmark_scenario.py
```

### NuPoCo Greedy

```
# source config_greedy.sh
# python run_benchmark_scenario.py
```

### NuPoCo

```
# source config_nupoco.sh
# python run_benchmark_scenario.py
```

Once the experiments have completed, use the following script to analyze the log files located in the *log* directory.
*log\_to\_csv\_scenario.py* generates three csv files, *ntt.csv*, *avg.csv*, *ttt.csv*, that analyze the performance with different metrics.

**ntt.csv**: normalized total turnaround times of co-located applications

**avg.csv**: average turnaround time of co-located applications

**ttt.csv**: total turnaround time (in absolute time) of co-located applications

Then, based on the csv file *gen\_plot.gnu* geneartes plots, *ntt.eps*, *avg.eps*, *ttt.eps* that correspond to Figure 11 in the paper.

```
$ python log_to_csv_scenario.py
$ ./gen_plot.gnu
or
$ gnuplot ./gen_plot.gnu
```

-------------------------------------------------------------------------------
## Customized Execution of the NuPoCo Resource Manager

First, execute the **nupoco_monitor** daemon:

```
sudo $RTE_ROOT/bin/nupoco_monitor
```

Second, execute OpenMP applications while linking the proper GOMP library and the communication library:

```
LD_PRELOAD="/path/to/snu-rte/lib/libRteComm.so" LD_LIBRARY_PATH="/path/to/the/GOMP/library/" OMP_SCHEDULE="rte" <executable path>
```



-------------------------------------------------------------------------------
## Paper

Younghyun Cho, Camilo A. Celis Guzman, and Bernhard Egger. "Maximizing System Utilization via Parallelism Management for Co-Located Parallel Applications", to appear in Proceedings of the international Conference on Parallel Architectures and Compilation Techniques (PACT), Limassol, Cyprus, 2018.

