#!/usr/bin/env python

import commands, subprocess, re, time, sys, csv
import datetime
from os.path import exists
import os
import itertools
import numpy
import csv
import operator

log_dir = "log/scenario/"

baseline_standalone_dir = "log/standalone/"

output_dir = "./"

number_re = "+([0-9]+\.[01-9]+|[0-9]+)"

apps = ["bt", "ep", "ft", "sp", "km"]

application_abbreviation = {"bt":"BT",\
                            "ep":"EP",\
                            "ft":"FT",\
                            "sp":"SP",\
                            "km":"KM"}

# benchmark parameters
scheduler_options = ["native","eqp","c3po","greedy","nupoco"]

def geometric_mean(iterable):
    return (reduce(operator.mul, iterable)) ** (1.0/len(iterable))

def get_baseline_turnaround_time(log):
    try:
        val = re.findall("Turnaround time: "+number_re, log)[0]
        return val
    except:
        return "-"

def get_turnaround_time(log, app):
    try:
        val = re.findall(app+"] Turnaround time: "+number_re, log)[0]
        return val
    except:
        return "-1.0"

if not exists(output_dir):
    os.system("mkdir -p "+output_dir)

scenario_num = 0

def analyze_log_2apps(app, ntt_csv_file, ttt_csv_file, average_ttt_csv_file):
    global scenario_num
    scenario_num += 1

    num_apps = len(app)

    baseline_ttt = []
    for iteration in ["1","2","3"]:
        logname_ = log_dir+"/native/"+"/native"
        for i in range(num_apps):
            logname_ += "."
            logname_ += app[i]
        logname_ += "."+iteration+".log"
        if exists(logname_):
            print logname_
            with open(logname_, "r") as f:
                log = f.read()
                max_app_tt = 0
                average_app_tt = 0
                for i in range(num_apps):
                    app_tt = float(get_turnaround_time(log, app[i]))
                    if app_tt > max_app_tt:
                        max_app_tt = app_tt
                    average_app_tt += app_tt
                baseline_ttt.append(max_app_tt)

    baseline_ttt_avg = numpy.sum(baseline_ttt)/len(baseline_ttt)

    batch_ntt = []
    batch_ttt = []
    for iteration in ["1","2","3"]:
        app_tt_sum = 0
        for i in range(num_apps):
            logname_ = baseline_standalone_dir+"/"+app[i]+".batch"+"."+iteration+".log"
            if exists(logname_):
                with open(logname_, "r") as f:
                    log = f.read()
                    app_tt_sum += float(get_baseline_turnaround_time(log))
        batch_ntt.append(app_tt_sum/baseline_ttt_avg)
        batch_ttt.append(app_tt_sum)

    batch_ntt_avg = numpy.sum(batch_ntt)/len(batch_ntt)
    batch_ntt_std = numpy.std(batch_ntt)

    batch_ttt_avg = numpy.sum(batch_ttt)/len(batch_ttt)
    batch_ttt_std = numpy.std(batch_ttt)

    output = ""
    for i in range(num_apps):
        if i < num_apps-1:
            output += application_abbreviation[app[i]] + "+"
        else:
            output += application_abbreviation[app[i]]
    ntt_csv_file.write(output + ", " + str(batch_ntt_avg) + ", " + str(batch_ntt_std))
    ttt_csv_file.write(output + ", " + str(batch_ttt_avg) + ", " + str(batch_ttt_std))
    average_ttt_csv_file.write(output)

    for scheduler in scheduler_options:
        log_name = scheduler
        print log_name

        for i in range(num_apps):
            log_name += "."
            log_name += app[i]

        scheduling_ttt = []
        scheduling_ntt = []
        scheduling_average = []

        for iteration in ["1","2","3"]:
            logname_ = log_dir+"/"+scheduler+"/"+log_name+"."+iteration+".log"
            if exists(logname_):
                with open(logname_, "r") as f:
                    log = f.read()
                    max_app_tt = 0
                    average_app_tt = 0
                    for i in range(num_apps):
                        app_tt = float(get_turnaround_time(log, app[i]))
                        if app_tt > max_app_tt:
                            max_app_tt = app_tt
                        average_app_tt += app_tt
                    scheduling_ttt.append(max_app_tt)
                    scheduling_ntt.append(max_app_tt/baseline_ttt_avg)
                    scheduling_average.append(average_app_tt/num_apps)

        scheduling_ttt_avg = numpy.sum(scheduling_ttt)/len(scheduling_ttt)
        scheduling_ttt_std = numpy.std(scheduling_ttt)

        scheduling_ntt_avg = numpy.sum(scheduling_ntt)/len(scheduling_ntt)
        scheduling_ntt_std = numpy.std(scheduling_ntt)

        scheduling_average_average = numpy.sum(scheduling_average)/len(scheduling_average)
        scheduling_average_std = numpy.std(scheduling_average)

        ntt_csv_file.write(", " + str(scheduling_ntt_avg) + ", " + str(scheduling_ntt_std))
        ttt_csv_file.write(", " + str(scheduling_ttt_avg) + ", " + str(scheduling_ttt_std))
        average_ttt_csv_file.write(", " + str(scheduling_average_average) + ", " + str(scheduling_average_std))

    ntt_csv_file.write("\n")
    ttt_csv_file.write("\n")
    average_ttt_csv_file.write("\n")

    return

def analyze_log_3apps(app, ntt_csv_file, ttt_csv_file, average_ttt_csv_file):
    global scenario_num
    scenario_num += 1

    num_apps = len(app)

    baseline_ttt = []
    for iteration in ["1","2","3"]:
        logname_ = log_dir+"/native/native"
        for i in range(num_apps):
            logname_ += "."
            logname_ += app[i]
        logname_ += "."+iteration+".log"
        if exists(logname_):
            print logname_
            with open(logname_, "r") as f:
                log = f.read()
                max_app_tt = 0
                average_app_tt = 0
                for i in range(num_apps):
                    app_tt = float(get_turnaround_time(log, app[i]))
                    if app_tt > max_app_tt:
                        max_app_tt = app_tt
                    average_app_tt += app_tt
                baseline_ttt.append(max_app_tt)

    baseline_ttt_avg = numpy.sum(baseline_ttt)/len(baseline_ttt)

    batch_ntt = []
    batch_ttt = []
    for iteration in ["1","2","3"]:
        app_tt_sum = 0
        for i in range(num_apps):
            logname_ = baseline_standalone_dir+"/"+app[i]+".batch"+"."+iteration+".log"
            if exists(logname_):
                with open(logname_, "r") as f:
                    log = f.read()
                    app_tt_sum += float(get_baseline_turnaround_time(log))
        batch_ntt.append(app_tt_sum/baseline_ttt_avg)
        batch_ttt.append(app_tt_sum)

    batch_ntt_avg = numpy.sum(batch_ntt)/len(batch_ntt)
    batch_ntt_std = numpy.std(batch_ntt)

    batch_ttt_avg = numpy.sum(batch_ttt)/len(batch_ttt)
    batch_ttt_std = numpy.std(batch_ttt)

    output = ""
    for i in range(num_apps):
        if i < num_apps-1:
            output += application_abbreviation[app[i]] + "+"
        else:
            output += application_abbreviation[app[i]]
    ntt_csv_file.write(output + ", " + str(batch_ntt_avg) + ", " + str(batch_ntt_std))
    ttt_csv_file.write(output + ", " + str(batch_ttt_avg) + ", " + str(batch_ttt_std))
    average_ttt_csv_file.write(output)

    for scheduler in scheduler_options:
        log_name = scheduler
        print log_name

        for i in range(num_apps):
            log_name += "."
            log_name += app[i]

        scheduling_ttt = []
        scheduling_ntt = []
        scheduling_average = []

        for iteration in ["1","2","3"]:
            logname_ = log_dir+"/"+scheduler+"/"+log_name+"."+iteration+".log"
            if exists(logname_):
                with open(logname_, "r") as f:
                    log = f.read()
                    max_app_tt = 0
                    average_app_tt = 0
                    for i in range(num_apps):
                        app_tt = float(get_turnaround_time(log, app[i]))
                        if app_tt > max_app_tt:
                            max_app_tt = app_tt
                        average_app_tt += app_tt
                    scheduling_ttt.append(max_app_tt)
                    scheduling_ntt.append(max_app_tt/baseline_ttt_avg)
                    scheduling_average.append(average_app_tt/num_apps)

        scheduling_ttt_avg = numpy.sum(scheduling_ttt)/len(scheduling_ttt)
        scheduling_ttt_std = numpy.std(scheduling_ttt)

        scheduling_ntt_avg = numpy.sum(scheduling_ntt)/len(scheduling_ntt)
        scheduling_ntt_std = numpy.std(scheduling_ntt)

        scheduling_average_average = numpy.sum(scheduling_average)/len(scheduling_average)
        scheduling_average_std = numpy.std(scheduling_average)

        ntt_csv_file.write(", " + str(scheduling_ntt_avg) + ", " + str(scheduling_ntt_std))
        ttt_csv_file.write(", " + str(scheduling_ttt_avg) + ", " + str(scheduling_ttt_std))
        average_ttt_csv_file.write(", " + str(scheduling_average_average) + ", " + str(scheduling_average_std))

    ntt_csv_file.write("\n")
    ttt_csv_file.write("\n")
    average_ttt_csv_file.write("\n")

    return

if not exists(output_dir):
  os.system("mkdir -p " + output_dir)

scenarios = [["bt","ft"],\
             ["bt","sp"],\
             ["ft","sp"],\
             ["bt","ft","sp"],\
             ["bt","ft","ep"],\
             ["bt","sp","ep"],\
             ["ft","sp","ep"],\
             ["bt","km"],\
             ["ft","km"],\
             ["sp","km"]]

scenario_num = 0
ntt_csv_file = open(output_dir+"/"+"ntt.csv","w+")
ttt_csv_file = open(output_dir+"/"+"ttt.csv","w+")
average_ttt_csv_file = open(output_dir+"/"+"avg.csv","w+")

ntt_csv_file.write("#bench, Batch, std, Native, std, EQP, std, NuPoCo Greedy, std, NuPoCo, std\n")
ttt_csv_file.write("#bench, Batch, std, Native, std, EQP, std, NuPoCo Greedy, std, NuPoCo, std\n")
average_ttt_csv_file.write("#bench, Native, std, EQP, std, NuPoCo Greedy, std, NuPoCo, std\n")

for app in scenarios:
    if len(app) == 2:
        analyze_log_2apps(app, ntt_csv_file, ttt_csv_file, average_ttt_csv_file)
    elif len(app) == 3:
        analyze_log_3apps(app, ntt_csv_file, ttt_csv_file, average_ttt_csv_file)
ntt_csv_file.close()
ttt_csv_file.close()
average_ttt_csv_file.close()

scheduler_average = []
scheduler_std = []

with open(output_dir+"/"+"ntt.csv","rw+") as f:
    f.readline()
    scheduler_scores = []
    scheduler_stds = []
    for i in range(len(scheduler_options)+1):
        scores = []
        scheduler_scores.append(scores)
        stds = []
        scheduler_stds.append(stds)
    for row in csv.reader(f):
        for i in range(len(scheduler_options)+1):
            avg_index = 1 + i * 2
            std_index = avg_index + 1
            scheduler_scores[i].append(float(row[avg_index]))
            scheduler_stds[i].append(float(row[std_index]))

    f.write("geomean")
    for i in range(len(scheduler_options)+1):
        try:
            f.write(", " + str(geometric_mean(scheduler_scores[i])) + ", " + str(geometric_mean(scheduler_stds[i])))
        except:
            f.write(", " + "0.0" + ", " + "0.0")
    f.write("\n")

    f.write("average")
    for i in range(len(scheduler_options)+1):
        try:
            f.write(", " + str(numpy.mean(scheduler_scores[i])) + ", " + str(geometric_mean(scheduler_stds[i])))
        except:
            f.write(", " + "0.0" + ", " + "0.0")
    f.write("\n")

with open(output_dir+"/"+"ttt.csv","rw+") as f:
    f.readline()
    scheduler_scores = []
    scheduler_stds = []
    for i in range(len(scheduler_options)+1):
        scores = []
        scheduler_scores.append(scores)
        stds = []
        scheduler_stds.append(stds)
    for row in csv.reader(f):
        for i in range(len(scheduler_options)+1):
            avg_index = 1 + i * 2
            std_index = avg_index + 1
            scheduler_scores[i].append(float(row[avg_index]))
            scheduler_stds[i].append(float(row[std_index]))

    f.write("geomean")
    for i in range(len(scheduler_options)+1):
        try:
            f.write(", " + str(geometric_mean(scheduler_scores[i])) + ", " + str(geometric_mean(scheduler_stds[i])))
        except:
            f.write(", " + "0.0" + ", " + "0.0")
    f.write("\n")

    f.write("average")
    for i in range(len(scheduler_options)+1):
        try:
            f.write(", " + str(numpy.mean(scheduler_scores[i])) + ", " + str(geometric_mean(scheduler_stds[i])))
        except:
            f.write(", " + "0.0" + ", " + "0.0")
    f.write("\n")

with open(output_dir+"/"+"avg.csv","rw+") as f:
    f.readline()
    scheduler_scores = []
    scheduler_stds = []
    for i in range(len(scheduler_options)):
        scores = []
        scheduler_scores.append(scores)
        stds = []
        scheduler_stds.append(stds)
    for row in csv.reader(f):
        for i in range(len(scheduler_options)):
            avg_index = 1 + i * 2
            std_index = avg_index + 1
            scheduler_scores[i].append(float(row[avg_index]))
            scheduler_stds[i].append(float(row[std_index]))

    f.write("geomean")
    for i in range(len(scheduler_options)):
        try:
            f.write(", " + str(geometric_mean(scheduler_scores[i])) + ", " + str(geometric_mean(scheduler_stds[i])))
        except:
            f.write(", " + "0.0" + ", " + "0.0")
    f.write("\n")

    f.write("average")
    for i in range(len(scheduler_options)):
        try:
            f.write(", " + str(numpy.mean(scheduler_scores[i])) + ", " + str(geometric_mean(scheduler_stds[i])))
        except:
            f.write(", " + "0.0" + ", " + "0.0")
    f.write("\n")

