#!/usr/bin/env python

import commands, subprocess, re, time, sys, csv
import datetime
from os.path import exists
import os
import itertools
import atexit

scheduler = os.environ["SCHEDULER_OPTION"]

log_dir = "log/scenario/"

binary = {"bt":"../applications/bin/bt.D.x",\
          "ep":"../applications/bin/ep.D.x",\
          "ft":"../applications/bin/ft.D.x",\
          "sp":"../applications/bin/sp.D.x",\
          "km":"../applications/kmeans/bin/kmeans"}

args = {"bt":" ",\
        "ep":" ",\
        "ft":" ",\
        "sp":" ",\
        "km":" -i ../applications/kmeans/data/inputGen/3000000_34.txt"}

@atexit.register
def clean_up():
  if rte_p:
    os.system("sudo kill -9 " + str(rte_p.pid))

def run_benchmark_2apps(scheduler, iter, app1, app2, logname):
    log = open(log_dir+"/"+scheduler+"/"+logname+".log", "w+")
    print "iter: " + iter + " app1: " + app1 + " app2: " + app2

    os.system("rm -f ready")

    # RTE enable
    global rte_p
    rte_p = None
    rte_root_path = os.environ["RTE_ROOT"]
    if rte_root_path is not "":
        rte_log = open(log_dir+"/"+scheduler+"/"+logname+".rte_log", "w+")
        print("Starting RTE")
        monitor_name = os.environ["RTE_MONITOR"]
        cmd = os.path.abspath(rte_root_path+"/bin/"+monitor_name)
        print(cmd)
        rte_p = subprocess.Popen(cmd.split(), stdout=rte_log, stderr=rte_log)

        # wait for RTE loading
        while True:
            if exists("ready"):
                break

        print("RTE started sucessfully")
        time.sleep(1)

    command_prefix = ""

    rte_comm_path = os.environ["RTE_COMM_PATH"]
    if rte_comm_path is not "":
        command_prefix += "LD_PRELOAD="+rte_comm_path+"/libRteComm.so "

    running_procs = []
    t_prev = datetime.datetime.now()

    print "start benchmark (current: " + str(t_prev) + ")"

    # execute application 1
    app_binary = " " + binary[app[0]]
    command = command_prefix + app_binary + " " + args[app[0]]

    print "app1: " + command
    p1 = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    running_procs.append(p1)

    # execute application 2
    app_binary = " " + binary[app[1]]
    command = command_prefix + app_binary + " " + args[app[1]]

    print "app2: " + command
    p2 = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    running_procs.append(p2)

    while running_procs:
        for proc in running_procs:
            retcode = proc.poll()
            if retcode is not None:
                running_procs.remove(proc)
                output = str(proc.communicate())

                turnaround_time = (datetime.datetime.now() - t_prev).total_seconds()

                if proc is p1:
                    app_name = app[0]
                elif proc is p2:
                    app_name = app[1]

                execution_log = open(log_dir+"/"+scheduler+"/execution_log/"+logname+"."+app_name+".execution.log","w+")
                execution_log.write(output)
                execution_log.close()

                finish_log              = "[" + app_name + "] Turnaround time: " + str(turnaround_time) + "\n"

                log.write(finish_log)

                break
            else:
                time.sleep(.1)
                continue

        if retcode != 0:
            """Error handling."""

    time.sleep(3)
    if rte_p:
        os.system("sudo kill -9 " + str(rte_p.pid))
        rte_p = None
        time.sleep(3)

    log.close()
    return

def run_benchmark_3apps(scheduler, iter, app1, app2, app3, logname):
    log = open(log_dir+"/"+scheduler+"/"+logname+".log", "w+")
    print "iter: " + iter + " app1: " + app1 + " app2: " + app2 + " app3: " + app3

    os.system("rm -f ready")

    # RTE enable
    global rte_p
    rte_p = None
    rte_root_path = os.environ["RTE_ROOT"]
    if rte_root_path is not "":
        rte_log = open(log_dir+"/"+scheduler+"/"+logname+".rte_log", "w+")
        print("Starting RTE")
        monitor_name = os.environ["RTE_MONITOR"]
        cmd = os.path.abspath(rte_root_path+"/bin/"+monitor_name)
        print(cmd)
        rte_p = subprocess.Popen(cmd.split(), stdout=rte_log, stderr=rte_log)

        # wait for RTE loading
        while True:
            if exists("ready"):
                break

        print("RTE started sucessfully")
        time.sleep(1)

    command_prefix = ""

    rte_comm_path = os.environ["RTE_COMM_PATH"]
    if rte_comm_path is not "":
        command_prefix += "LD_PRELOAD="+rte_comm_path+"/libRteComm.so "

    running_procs = []
    t_prev = datetime.datetime.now()

    print "start benchmark (current: " + str(t_prev) + ")"

    # execute application 1
    app_binary = " " + binary[app[0]]
    command = command_prefix + app_binary + " " + args[app[0]]

    print "app1: " + command
    p1 = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    running_procs.append(p1)

    # execute application 2
    app_binary = " " + binary[app[1]]
    command = command_prefix + app_binary + " " + args[app[1]]

    print "app2: " + command
    p2 = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    running_procs.append(p2)

    # execute application 3
    app_binary = " " + binary[app[2]]
    command = command_prefix + app_binary + " " + args[app[2]]

    print "app3: " + command
    p3 = subprocess.Popen(command, shell=True, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    running_procs.append(p3)

    while running_procs:
        for proc in running_procs:
            retcode = proc.poll()
            if retcode is not None:
                running_procs.remove(proc)
                output = str(proc.communicate())

                turnaround_time = (datetime.datetime.now() - t_prev).total_seconds()

                if proc is p1:
                    app_name = app[0]
                elif proc is p2:
                    app_name = app[1]
                elif proc is p3:
                    app_name = app[2]

                execution_log = open(log_dir+"/"+scheduler+"/execution_log/"+logname+"."+app_name+".execution.log","w+")
                execution_log.write(output)
                execution_log.close()

                finish_log              = "[" + app_name + "] Turnaround time: " + str(turnaround_time) + "\n"

                log.write(finish_log)

                break
            else:
                time.sleep(.1)
                continue

        if retcode != 0:
            """Error handling."""

    time.sleep(3)
    if rte_p:
        os.system("sudo kill -9 " + str(rte_p.pid))
        rte_p = None
        time.sleep(3)

    log.close()
    return

if not exists(log_dir+"/"+scheduler):
    os.system("mkdir -p "+log_dir)
    os.system("mkdir -p "+log_dir+"/"+scheduler)
    os.system("mkdir -p "+log_dir+"/"+scheduler+"/"+"execution_log")

scenarios = [["bt","ft"],\
             ["bt","sp"],\
             ["ft","sp"],\
             ["bt","ft","sp"],\
             ["bt","ft","ep"],\
             ["bt","sp","ep"],\
             ["ft","sp","ep"],\
             ["bt","km"],\
             ["ft","km"],\
             ["sp","km"]]

for iter in ["1","2","3"]:
    for app in scenarios:
        if len(app) == 2:
            logname = scheduler+"."+app[0]+"."+app[1]+"."+iter

            if exists(log_dir+"/"+scheduler+"/"+logname+".log"):
                print logname + ".log" + " is already exists"
            else:
                time.sleep(3)
                run_benchmark_2apps(scheduler, iter, app[0], app[1], logname)
        elif len(app) == 3:
            logname = scheduler+"."+app[0]+"."+app[1]+"."+app[2] +"."+iter

            if exists(log_dir+"/"+scheduler+"/"+logname+".log"):
                print logname + ".log" + " is already exists"
            else:
                time.sleep(3)
                run_benchmark_3apps(scheduler, iter, app[0], app[1], app[2], logname)
