//------------------------------------------------------------------------------
/// @brief ARM - RTE communication library
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
/// 
/// @decription rte_comm.h
/// ARM - RTE communication library header
/// 
//------------------------------------------------------------------------------

#ifndef __RTE_COMM_H_
#define __RTE_COMM_H_

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <dlfcn.h>
#include "comm_interface.h"

#include <numa.h>

#define DEBUG_MODE    ///< debug mode shows debug messages

int    num_cores;

int    shmid;      ///< shared memory area id
key_t  key;        ///< shared memory area key (application's pid)
void*  shmaddr;    ///< shared memory area address
int    shm_opened; ///< shared memory initialization flag

/* main initialization routine interpositioning */
int __libc_start_main(int (*main) (int, char **, char **),
    int argc,
    char ** ubp_av,
    void (*init) (void),
    void (*fini) (void),
    void (*rtld_fini) (void),
    void (*stack_end));

/* exit call interpositioning */
void exit(int status);

/* initialize shared memory area */
void InitSHM(void);

/* cleanup shared memory area */
void CleanUpSHM(void);

/* StartApplication implementation (APP_BEGIN) */
int RTE_StartApplication(char* buf);

/* FinishApplication implementation (APP_END) */
int RTE_FinishApplication(void);

#endif // __RTE_COMM_H_

