//------------------------------------------------------------------------------
/// @brief ARM - RTE misc functions
/// @author Camilo Celis Guzman <camilo@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription misc.cpp
/// set of helpful misc functions
///
//------------------------------------------------------------------------------

#include "misc.h"

using namespace std;

//------------------------------------------------------------------------------
void print_backtrace(size_t backtrace_depth, struct sigcontext context)
{
  void** trace = (void**)calloc(sizeof(void*), backtrace_depth);
  int trace_size = 0;
  trace_size = backtrace(trace, backtrace_depth);

  // define stack pointer
#if defined(__i386__)
  trace[1] = (void*)context.eip;
#elif defined(__x86_64__)
  trace[1] = (void*)context.rip;
#else
#error Unsupported architecture
#endif

  // print trace
  cout << "\n" + string(80, '-') << endl << "Execution path:" << endl;
  // from i = 1, ignore signal handler function call
  for(int i = 2; i < trace_size; ++i)
  {
    // get executable path
    char exe[1024];
    if(readlink("/proc/self/exe", exe, sizeof(exe) - 1) == -1)
    {
      cout << "readlink error" << endl;
      free(trace);
      exit(1);
    }

    char result[1024] = {0,};
    // use addr2line to read the file and line number
    // TODO: Check if add2line is present
    char buf[256];
    sprintf(buf, "addr2line %p -e %s 2>/dev/null", trace[i], exe);
    FILE* fd = popen(buf, "r");
    size_t read_size = fread(result, sizeof(char), sizeof(result) - 1, fd);
    if(read_size != sizeof(result) - 1)
      cerr << "Error reading reading buffer from addr2line" << endl;
    pclose(fd);

    cout << "[trace] " << setw(3) << i << ": " << result;
  }
  free(trace);
}
