//------------------------------------------------------------------------------
/// @brief ARM - RTE misc functions
/// @author Camilo Celis Guzman <camilo@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription misc.cpp
/// set of helpful misc functions
///
//------------------------------------------------------------------------------

#ifndef __MISC_H__
#define __MISC_H__

#include <iostream>     ///< cout, endl
#include <iomanip>      ///< setw
#include <stdlib.h>     ///< calloc
#include <signal.h>     ///< sigcontext
#include <execinfo.h>   ///< backtrace
#include <unistd.h>     ///< readlink

#include "common.h"

//------------------------------------------------------------------------------
// Macros
#define unlikely(x)   __builtin_expect(static_cast<bool>(x), 1)
#define likely(x)     __builtin_expect(static_cast<bool>(x), 0)

#define BACKTRACE_DEPTH      32   ///< depth use during backtrace

//------------------------------------------------------------------------------
/// @brief Writes to stdout a backtrace result of given size for current process
/// @param backtrace_depth backtrace depth
/// @param signal context to get the stack pointer
void print_backtrace(size_t backtrace_depth, struct sigcontext context);

#endif /* __MISC_H__ */



