#ifndef __PERFORMANCE_MODEL_
#define __PERFORMANCE_MODEL_

#include <math.h>

inline double compute_mm1_model(int tiles, double service_rate,
    double llc_all, int memory_controllers, int max_tiles)
{
    double service_time, max_service_time;

    service_time =
        1.0 / (service_rate - (llc_all/(double)memory_controllers));

    max_service_time =
        100.0 * double(max_tiles) / double(memory_controllers);

    if (service_time < 0 || service_time >= max_service_time) {
        return max_service_time;
    }

    return service_time;
}

inline double prob_nothing_in_the_system(int num_customers, double rho)
{
    double ret = 1.0;

    for (int k = 1; k <= num_customers; k++) {
        double each_prob = 1.0;
        int mult = num_customers;

        for (int j = 0; j < k; j++) {
            each_prob = each_prob * (double)mult;
            mult = mult - 1;
        }

        each_prob = each_prob * pow(rho, k);
        ret = ret + each_prob;
    }

    return (1.0 / ret);
}

inline double mm1nn_service_time(double service_rate,
    double arrival_rate_from_one_customer,
    int num_customers)
{
    double rho = arrival_rate_from_one_customer / service_rate;
    double prob_something_in_the_system =
        1.0 - prob_nothing_in_the_system(num_customers, rho);

    return (1.0 / service_rate) *
        (((double)num_customers / prob_something_in_the_system) - (1.0 / rho));
}

inline double mm1nn_server_utilization(
    double service_rate,
    double arrival_rate_from_one_customer,
    double num_customers)
{
    double rho = arrival_rate_from_one_customer / service_rate;
    double prob_something_in_the_system =
        1.0 - prob_nothing_in_the_system(num_customers, rho);

    return prob_something_in_the_system;
    //return (1.0 / service_rate) *
    //    (((double)num_customers / prob_something_in_the_system) - (1.0 / rho));
}

#endif
