//------------------------------------------------------------------------------
/// @brief ARM - RTE prototype instance
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription prototype.cpp
/// Basic implementation of the Many-SC prototype.
///
//------------------------------------------------------------------------------

#include "prototype.h"

//------------------------------------------------------------------------------
CPrototype::CPrototype()
{
}

CPrototype::~CPrototype()
{
}

void CPrototype::BriefSystemInfo(CPUArchitecture* arch,
    CoreAllocator* scheduler, CProfiler* profiler, string path)
{
  ofstream out(path.c_str());

  out << "========================================"
      << "========================================" << endl;
  out << "Brief System Information" << endl;

  arch->BriefArchitectureInfo(path);
  scheduler->SchedulerInfo(path);
  profiler->ProfilerInfo(path);

  out << endl;

  out.close();
}

void CPrototype::DetailedSystemInfo(CPUArchitecture* arch,
    CoreAllocator* scheduler, CProfiler* profiler, string path)
{
  ofstream out(path.c_str());

  out << "========================================"
      << "========================================" << endl;
  out << "Detailed System Information" << endl;

  if (arch != NULL)
    arch->DetailedArchitectureInfo(path);
  if (scheduler != NULL)
    scheduler->SchedulerInfo(path);
  if (profiler != NULL)
    profiler->ProfilerInfo(path);

  out << endl;

  out.close();
}

CPUArchitecture* CPrototype::CreateArchitectureInstance()
{
    CPUArchitecture* arch = new CPUArchitecture();

    return arch;
}

CoreAllocator* CPrototype::CreateScheduler(CPUArchitecture* arch)
{
  CoreAllocator* scheduler = new CoreAllocator(arch);
  scheduler->SetPolicy(EQUAL_PARTITION);

  return scheduler;
}

CoreAllocator* CPrototype::CreateEqualPartitionScheduler(CPUArchitecture* arch)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(EQUAL_PARTITION);

    return scheduler;
}

CoreAllocator* CPrototype::CreateInterleavingScheduler(CPUArchitecture* arch)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(INTERLEAVING);

    return scheduler;
}

CoreAllocator* CPrototype::
CreateMaximizeSpeedupScheduler(CPUArchitecture* arch)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(MAX_SPEEDUP);

    return scheduler;
}

CoreAllocator* CPrototype::
CreateMaximizeUtilizationScheduler(CPUArchitecture* arch)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(MAX_UTILIZATION);

    return scheduler;
}

CoreAllocator* CPrototype::
CreateHeuristicsScheduler(CPUArchitecture* arch)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(HEURISTICS);

    return scheduler;
}

CoreAllocator* CPrototype::
CreateGreedyScheduler(CPUArchitecture* arch)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(GREEDY);

    return scheduler;
}

CoreAllocator* CPrototype::
CreateLocalityAwareAllocator(CPUArchitecture* arch,
        int num_cores_per_node)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(LOCALITY_AWARE);
    scheduler->SetNumCoresPerNode(num_cores_per_node);

    return scheduler;
}

CoreAllocator* CPrototype::
CreateSimpleAllocator(CPUArchitecture* arch,
        int num_allocated_cores)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(SIMPLE);
    scheduler->SetNumCores(num_allocated_cores);

    return scheduler;
}

CoreAllocator* CPrototype::CreateC3POScheduler(CPUArchitecture* arch)
{
    CoreAllocator* scheduler = new CoreAllocator(arch);
    scheduler->SetPolicy(C3PO);

    return scheduler;
}

CProfiler* CPrototype::CreateSimpleProfiler(CPUArchitecture* arch,
    CoreAllocator* scheduler)
{
  CProfiler* profiler = new CSimpleProfiler(arch, scheduler);

  return profiler;
}

CProfiler* CPrototype::CreateHeuristicsProfiler(CPUArchitecture* arch,
    CoreAllocator* scheduler)
{
    CProfiler* profiler = new CHeuristicsProfiler(arch, scheduler);

    return profiler;
}

CProfiler* CPrototype::CreateC3POProfiler(CPUArchitecture* arch,
    CoreAllocator* scheduler)
{
  CProfiler* profiler = new CC3POProfiler(arch, scheduler);

  return profiler;
}

RTEAppConnection* CPrototype::CreateApplicationSocketConnection()
{
  RTEAppConnection* app_connection = new RTEAppConnection(11111);

  return app_connection;
}

