//------------------------------------------------------------------------------
/// @brief RTE CPU description
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
/// 
//------------------------------------------------------------------------------

#ifndef __RTE_CPU_H_
#define __RTE_CPU_H_

/// @brief rte_cpu_set declaration
///
/// @description we currently follow the same interface with
///              Linux cpu_set_t structure (include/linux/sched.h)
///              this description is for quick implementation.
///              (to maintain same structure with sc_shepherd kernel module)
///              it will be improved to use better model (e.g., use cpu_set_t)
typedef struct
{
  unsigned long bits; //maximum cpu number is 64
} rte_cpu_set;

/// @brief CPU_ZERO()
///
/// @description clear set, so that it contains no cpu set
static void RTE_CPU_ZERO(rte_cpu_set *set)
{
  set->bits = 0x00UL;
}

/// @brief CPU_SET()
///
/// @description add cpu to set
static void RTE_CPU_SET(int cpu, rte_cpu_set *set)
{
  set->bits |= ((0x01UL) << cpu);
}

/// @brief CPU_CLR()
///
/// @description remove cpu from set
static void RTE_CPU_CLR(int cpu, rte_cpu_set *set)
{
  set->bits &= (~((0x01UL) << cpu));
}

/// @brief CPU_ISSET
///
/// @description test to see if cpu is a member of set
static int RTE_CPU_ISSET(int cpu, rte_cpu_set *set)
{
  if (set->bits & (0x01UL << cpu)) {
    return 1;
  } else {
    return 0;
  }
}

#endif // __RTE_CPU_H_
